package com.me.pacman.state;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Input;
import com.badlogic.gdx.InputAdapter;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.g2d.BitmapFont;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.Vector2;
import com.me.pacman.PacDude;
import com.me.pacman.Score;
import com.me.pacman.Sound;
import com.me.pacman.entity.*;
import com.me.pacman.entity.ai.ReturnToBase;
import com.me.pacman.entity.path.EnterGhostHousePath;
import com.me.pacman.level.Level;
import com.me.pacman.level.LevelTile;
import com.me.pacman.level.Modifiers;

import java.util.Random;

public class PlayState extends LevelState {

    public static final float FULL_SPEED = 9.375f;

    public static final Vector2[] GHOST_SPAWN_POINTS = {
        new Vector2(14f, 19.5f),
        new Vector2(14f, 16.5f),
        new Vector2(12f, 16.5f),
        new Vector2(16f, 16.5f),
    };

    public static final Direction[] GHOST_SPAWN_DIRS = {
            Direction.LEFT,
            Direction.UP,
            Direction.DOWN,
            Direction.DOWN,
    };

    private Texture levelBackground, winBackground;
    private TextureRegion lifeSprite;

    public Random random;

    public int pelletsRemaining;
    private int pelletsEaten;

    public int pelletsEatenSinceDeath;
    public boolean pelletsEatenSinceDeathCounterEnabled;
    public boolean hasDied;

    private int score;
    private int highScore;
    private int lives;
    public int round;
    private boolean paused = false;

    public float secondsSinceLastDot;

    public boolean scatter;
    private int scatterChaseTransition;
    private float scatterChaseTimer;

    public float frightTimer;           // remaining fright time
    private int ghostsCaught;           // number of ghosts caught since last power pellet
    private Ghost lastGhostCaptured;

    public Pacman pacman;
    public Ghost[] ghosts;

    public enum GameState {
        PRE_NEW_GAME(2.3f),
        NEW_ROUND_WAIT(4.2f),
        START_ROUND_WAIT(2f),
        ROUND_WON_WAIT(1f),
        ROUND_WON(2f),
        PACMAN_CAUGHT,
        PACMAN_CAUGHT_WAIT(1f),
        GHOST_CAUGHT_POINTS_WAIT(1f),
        GAME_OVER(2f),
        PLAYING,
        ;

        final float timer;

        GameState(float timer) {
            this.timer = timer;
        }

        GameState() {
            this(0f);
        }
    }

    private GameState state;
    private float stateTimer;

    public PlayState(PacDude game) {
        super(game);
    }

    public void updateBackgroundAudio() {
        for (Ghost ghost : ghosts) {
            if ((ghost.currentBehaviour instanceof ReturnToBase
                    || ghost.currentPath instanceof EnterGhostHousePath) && !ghost.inHouse) {
                if (!game.sound.isLooping(Sound.Effect.RETURN_BASE)) {
                    game.sound.stopLoops();
                    game.sound.loop(Sound.Effect.RETURN_BASE);
                }
                return;
            }
        }

        if (frightTimer > 0) {
            if (!game.sound.isLooping(Sound.Effect.FRIGHT)) {
                game.sound.stopLoops();
                game.sound.loop(Sound.Effect.FRIGHT);
            }
            return;
        }

        if (!game.sound.isLooping(Sound.Effect.SIREN)) {
            game.sound.stopLoops();
            game.sound.loop(Sound.Effect.SIREN);
        }
    }

    private void setGameState(GameState state) {
        if (state == null) return;
        this.state = state;
        this.stateTimer = state.timer;
    }

    private void spawnGhosts() {
        ghosts = new Ghost[4];
        ghosts[0] = new Blinky(this, new Vector2(GHOST_SPAWN_POINTS[0]), GHOST_SPAWN_DIRS[0]);
        ghosts[1] = new Pinky(this, new Vector2(GHOST_SPAWN_POINTS[1]), GHOST_SPAWN_DIRS[1]);
        ghosts[2] = new Inky(this, new Vector2(GHOST_SPAWN_POINTS[2]), GHOST_SPAWN_DIRS[2]);
        ghosts[3] = new Clyde(this, new Vector2(GHOST_SPAWN_POINTS[3]), GHOST_SPAWN_DIRS[3]);
    }

    public void resetGhosts() {
        for (int i = 0; i < 4; i++) {
            if (ghosts[i] == null) continue;
            ghosts[i].pos = new Vector2(GHOST_SPAWN_POINTS[i]);
            ghosts[i].currDirection = GHOST_SPAWN_DIRS[i];
            ghosts[i].currentPath = null;
            ghosts[i].caught = false;
            ghosts[i].inHouse = i > 0;
            ghosts[i].updateBehaviour();
        }
    }

    private void initializeLevel() {
        level = new Level(game,"level");
        pelletsRemaining = level.getTileCount(LevelTile.PELLET) + level.getTileCount(LevelTile.POWER_PELLET);
        pacman = new Pacman(this, false);
    }

    private void initializeRound() {
        scatter = true;
        scatterChaseTransition = 0;

        pelletsEaten = 0;
        pelletsEatenSinceDeath = 0;
        pelletsEatenSinceDeathCounterEnabled = false;
        hasDied = false;

        spawnGhosts();
    }

    private void resetField() {
        frightTimer = 0f;
        ghostsCaught = 0;
        lastGhostCaptured = null;
        secondsSinceLastDot = 0;

        scatterChaseTimer = Modifiers.getScatterChaseTimer(round, scatterChaseTransition);

        random = new Random(897198256012865L);

        pacman = new Pacman(this, false);
        resetGhosts();
    }

    private void preNewGame() {
        score = 0;

        Score hs = game.highScores.getCurrentHighScore();
        highScore = hs == null ? 0 : hs.score;

        lives = 3;
        round = 0;

        initializeLevel();

        game.sound.stopAll();
        game.sound.play(Sound.Effect.BEGINNING);
    }

    private void newGame() {
        lives--;
        initializeRound();
        resetField();
    }

    private void newRound() {
        initializeLevel();
        initializeRound();
        resetField();

        round++;

        game.sound.stopAll();
        game.sound.play(Sound.Effect.BEGINNING_ALT);
    }

    private void startGame() {
        game.sound.loop(Sound.Effect.SIREN);
        pacman.moving = true;
    }

    @Override
    public void setup() {
        levelBackground = game.assets.getLevelBackground();
        winBackground = game.assets.getLevelWinBackground();
        lifeSprite = game.assets.pacman[2][1];
        Gdx.input.setInputProcessor(new Controller());

        preNewGame();
        setGameState(GameState.PRE_NEW_GAME);
    }

    public GameState stateTransition() {
        switch (state) {
            // The state we're transitioning /from/
            case PRE_NEW_GAME:
                newGame();
                return GameState.START_ROUND_WAIT;
            case START_ROUND_WAIT:
            case NEW_ROUND_WAIT:
                startGame();
                return GameState.PLAYING;
            case ROUND_WON_WAIT:
                return GameState.ROUND_WON;
            case ROUND_WON:
                newRound();
                return GameState.NEW_ROUND_WAIT;
            case GAME_OVER:
                if (score > highScore) {
                    game.setNextState(new HighScoreEntryState(game, new Score("aaa", score)));
                    return null;
                }
                preNewGame();
                return GameState.PRE_NEW_GAME;
            case GHOST_CAUGHT_POINTS_WAIT:
                updateBackgroundAudio();
                return GameState.PLAYING;
            case PACMAN_CAUGHT_WAIT:
                game.sound.play(Sound.Effect.DEATH);
                pacman.alive = false;
                return GameState.PACMAN_CAUGHT;
        }
        return null;
    }

    @Override
    public void render() {

        game.fontRenderer.setColor(Color.WHITE);
        game.fontRenderer.draw(game.batch, "1up", 3 * 8, 35 * 8);

        // Draw score
        // Determine x position based on score size.
        String s = score > 0 ? Integer.toString(score) : "00";
        game.fontRenderer.draw(game.batch, s, (7 - s.length()) * 8, 34 * 8);

        // Draw high score
        game.fontRenderer.draw(game.batch, "high score", 9 * 8, 35 * 8);
        game.fontRenderer.draw(game.batch, score >= highScore? s : Integer.toString(highScore), 12 * 8, 34 * 8);

        // Draw remaining lives
        for (int i = 0; i < lives; i++) {
            game.batch.draw(lifeSprite, i * 16, 0);
        }

        // Draw the level tiles
        level.render(LEVEL_OFFSET_X, LEVEL_OFFSET_Y);

        if (state == GameState.ROUND_WON) {
            // draw flashing level background
            game.batch.draw((int) (stateTimer * 4) % 2 == 0? levelBackground : winBackground, LEVEL_OFFSET_X, LEVEL_OFFSET_Y);
            return;
        } else {
            game.batch.draw(levelBackground, LEVEL_OFFSET_X, LEVEL_OFFSET_Y);
        }

        if (state != GameState.GHOST_CAUGHT_POINTS_WAIT) {
            pacman.render();
        }

        switch (state) {
            case PRE_NEW_GAME:
            case PACMAN_CAUGHT:
            case GAME_OVER:
                break;
            case GHOST_CAUGHT_POINTS_WAIT:
                drawSprite(game.assets.points[0][ghostsCaught-1], lastGhostCaptured.pos.x, lastGhostCaptured.pos.y);
            default:
                for (Ghost ghost : ghosts) {
                    if (state == GameState.GHOST_CAUGHT_POINTS_WAIT && ghost == lastGhostCaptured) {
                        continue;
                    }
                    ghost.render();
                }
        }

        if (paused) {
            game.fontRenderer.setColor(Color.YELLOW);
            game.fontRenderer.draw(game.batch, "paused", 11 * 8, 15 * 8);
        } else {
            switch (state) {
                case PRE_NEW_GAME:
                case NEW_ROUND_WAIT:
                case START_ROUND_WAIT:
                    game.fontRenderer.setColor(Color.YELLOW);
                    game.fontRenderer.draw(game.batch, "ready!", 11 * 8, 15 * 8);
                    break;
                case GAME_OVER:
                    game.fontRenderer.setColor(Color.RED);
                    game.fontRenderer.draw(game.batch, "game over", (9 * 8) + 4, 15 * 8);
                    break;
            }
        }
    }

    private void pelletEaten(float x, float y) {
        level.setTile(x, y, LevelTile.EMPTY);

        game.sound.play(pelletsEaten % 2 == 0? Sound.Effect.CHOMP_1 : Sound.Effect.CHOMP_2);

        pelletsEaten++;
        pelletsRemaining--;

        if (pelletsEatenSinceDeathCounterEnabled) {
            // Increase global dot counter when enabled
            pelletsEatenSinceDeath++;
        } else {
            // else increment appropriate ghost's individual counter
            for (Ghost ghost : ghosts) {
                // Increment the dot counter of the first encountered ghost in the house
                // Ghosts are ordered Blinky, Pinky, Inky, Clyde. Blinky's inHouse is always false,
                // Increase the dotCounter of only one ghost: the first of the above list still in the house.
                if (!ghost.inHouse) continue;
                ghost.pelletCounter++;
                break;
            }
        }

        secondsSinceLastDot = 0;

        if (pelletsRemaining == 0) {
            game.sound.stopLoops();
            setGameState(GameState.ROUND_WON_WAIT);
        }
    }

    public void eatPellet(float x, float y) {
        pelletEaten(x, y);
        score += 10;
    }

    public void eatPowerPellet(float x, float y) {
        pelletEaten(x, y);
        for (Ghost ghost : ghosts) {
            if (ghost.currentBehaviour instanceof ReturnToBase) continue;
            ghost.caught = false;
            ghost.currentBehaviour = ghost.frightBehaviour;
            ghost.reverse = true;
        }
        frightTimer = Modifiers.getFrightTime(round);
        ghostsCaught = 0;
        score += 50;

        updateBackgroundAudio();
    }

    private void pacmanCaught() {
        game.sound.stopLooping(Sound.Effect.SIREN);
        pacman.moving = false;
        pelletsEatenSinceDeath = 0;
        pelletsEatenSinceDeathCounterEnabled = true;
        hasDied = true;

        setGameState(GameState.PACMAN_CAUGHT_WAIT);
    }

    private void ghostCaught(Ghost ghost) {
        ghost.caught = true;
        ghost.currentBehaviour = new ReturnToBase(this);

        lastGhostCaptured = ghost;
        ghostsCaught++;
        score += ghostsCaught * 200;

        game.sound.play(Sound.Effect.EAT_GHOST);
        setGameState(GameState.GHOST_CAUGHT_POINTS_WAIT);
    }

    private void updateScatterTimer(float dt) {
        if (scatterChaseTransition < 6) {
            scatterChaseTimer -= dt;
            if (scatterChaseTimer <= 0) {
                scatter = !scatter;
                scatterChaseTransition++;
                scatterChaseTimer = Modifiers.getScatterChaseTimer(round, scatterChaseTransition);
            }
        }
    }

    private void updateFrightTimer(float dt) {
        frightTimer -= dt;
        if (frightTimer <= 0) {
            for (Ghost ghost : ghosts) {
                if (ghost.currentBehaviour instanceof ReturnToBase) continue;
                ghost.caught = false;
                ghost.currentBehaviour = ghost.chaseBehaviour;
            }

            updateBackgroundAudio();
        }
    }

    private void updateSecondsSinceLastDot(float dt) {
        secondsSinceLastDot += dt;
        if (secondsSinceLastDot >= Modifiers.getForceLeaveSeconds(round)) {
            // It's been 4 seconds since pacman last ate a dot, he's tryin' to avoid ghosts coming out!
            // We'll get him...
            for (int i = 1; i < 4; i++) {
                Ghost ghost = ghosts[i];
                if (ghost.inHouse) {
                    ghost.leaveHouse();
                    secondsSinceLastDot = 0;
                    break;
                }
            }
        }
    }

    // Returns whether pacman has been caught
    private boolean handleGhostCollision(Ghost ghost) {
        if (ghost.onSameTile(pacman)) {
            if (frightTimer > 0 && !ghost.caught) {
                ghostCaught(ghost);
            } else if (!(ghost.currentBehaviour instanceof ReturnToBase)) {
                pacmanCaught();
            }
            return true;
        }
        return false;
    }

    @Override
    public void update(float dt) {
        if (paused) {
            return;
        }

        if (PacDude.DEBUG) {
            // Fixed time step for debugger
            dt = 1/60f;
        }

        switch (state) {
            case PLAYING:
                if (frightTimer <= 0) {
                    updateScatterTimer(dt);
                } else {
                    updateFrightTimer(dt);
                }

                updateSecondsSinceLastDot(dt);

                pacman.update(dt);
                for (Ghost ghost : ghosts) {
                    ghost.update(dt);
                    if (handleGhostCollision(ghost)) {
                        return;
                    }
                }
                break;
            case PACMAN_CAUGHT:
                pacman.update(dt);
                if (pacman.deathFrame == 13) {
                    if (lives > 0) {
                        lives--;
                        resetField();
                        setGameState(GameState.START_ROUND_WAIT);
                    } else {
                        setGameState(GameState.GAME_OVER);
                    }
                }
                break;
            case GHOST_CAUGHT_POINTS_WAIT:
                for (Ghost ghost : ghosts) {
                    if (ghost.currentBehaviour instanceof ReturnToBase && ghost != lastGhostCaptured) {
                        ghost.update(dt);
                    }
                }
                break;
        }

        if (state.timer > 0) {
            stateTimer -= dt;
            if (stateTimer <= 0) {
                setGameState(stateTransition());
            }
        }
    }

    private final class Controller extends InputAdapter {

        int downX;
        int downY;
        long downTime;

        @Override
        public boolean keyDown(int keycode) {
            switch (keycode) {
                case Input.Keys.UP:
                case Input.Keys.W:
                    pacman.setNextDirection(Direction.UP);
                    break;
                case Input.Keys.DOWN:
                case Input.Keys.S:
                    pacman.setNextDirection(Direction.DOWN);
                    break;
                case Input.Keys.LEFT:
                case Input.Keys.A:
                    pacman.setNextDirection(Direction.LEFT);
                    break;
                case Input.Keys.RIGHT:
                case Input.Keys.D:
                    pacman.setNextDirection(Direction.RIGHT);
                    break;
                case Input.Keys.ESCAPE:
                case Input.Keys.P:
                    paused = !paused;
                    if (paused) {
                        game.sound.pauseAll();
                    } else {
                        game.sound.resumeAll();
                    }
                    break;
                case Input.Keys.N:
                    preNewGame();
                    setGameState(GameState.PRE_NEW_GAME);
                    break;
            }
            return super.keyDown(keycode);
        }

        @Override
        public boolean touchDown(int screenX, int screenY, int pointer, int button) {
            if (Math.abs(downX - screenX) < 50 && Math.abs(downY - screenY) < 50
                    && System.currentTimeMillis() - downTime < 250) {
                paused = !paused;
                return true;
            }
            downX = screenX;
            downY = screenY;
            downTime = System.currentTimeMillis();
            return true;
        }

        @Override
        public boolean touchUp(int screenX, int screenY, int pointer, int button) {
            int relaX = screenX - downX;
            int relaY = screenY - downY;

            if (Math.abs(relaX) < 50 && Math.abs(relaY) < 50) {
                // didn't move enough to consider a swipe
                return true;
            }

            if (Math.abs(relaX) > Math.abs(relaY)) {
                // x > y, so moving left-right
                if (relaX > 0) {
                    pacman.setNextDirection(Direction.RIGHT);
                } else {
                    pacman.setNextDirection(Direction.LEFT);
                }
            } else {
                // else, moving up/down
                if (relaY > 0) {
                    pacman.setNextDirection(Direction.DOWN);
                } else {
                    pacman.setNextDirection(Direction.UP);
                }
            }
            return true;
        }

    }

    @Override
    public void dispose() {
        Gdx.input.setInputProcessor(null);
    }

}
